import '../io_driver/io_driver_browser.dart';
import 'package:universal_io/src/io/all.dart';
import 'chrome/chrome_common.dart';
import 'chrome/chrome_sockets.dart';
import 'dart:collection';
import 'dart:async';
import 'package:raw/raw.dart';
import 'dart:typed_data';

class ChromeIODriver extends IODriver {
  @override
  Future<RawSocket> connectRawSocket(host, int port,
      {sourceAddress, Duration timeout}) async {
    final createInfo = await sockets.tcp.create();
    final socketId = createInfo.socketId;
    try {
      final resultValue =
          await sockets.tcp.connect(socketId, host.toString(), port);
      if (resultValue < 0) {
        throw StateError(
            "Creating TCP connection from '$sourceAddress' to '$host:$port' failed with error code $resultValue");
      }
      return ChromeRawSocket(
        socketId,
        remoteAddress: InternetAddress(host),
        remotePort: port,
      );
    } catch (e) {
      sockets.udp.close(socketId);
      rethrow;
    }
  }

  @override
  Future<RawDatagramSocket> bindRawDatagramSocket(host, int port,
      {bool reuseAddress = true, bool reusePort = false, int ttl = 1}) async {
    final createInfo = await sockets.udp.create();
    final socketId = createInfo.socketId;
    try {
      final resultValue =
          await sockets.udp.bind(socketId, host.toString(), port ?? 0);
      if (resultValue < 0) {
        throw StateError(
          "Binding UDP socket to $host:$port failed with error code $resultValue",
        );
      }
      return ChromeRawDatagramSocket(
        socketId,
        address: InternetAddress(host),
        port: port,
      );
    } catch (e) {
      sockets.udp.close(socketId);
      rethrow;
    }
  }
}

class ChromeRawSocket extends Stream<RawSocketEvent> implements RawSocket {
  final int socketId;
  final StreamController<RawSocketEvent> _eventStreamController =
      StreamController<RawSocketEvent>();
  RawWriter _incomingWriter = RawWriter.withCapacity(128);

  ChromeRawSocket(this.socketId,
      {this.address, this.port, this.remotePort, this.remoteAddress}) {
    _receivingSockets[socketId] = this;
    _startReceiving();
  }

  static Map<int, ChromeRawSocket> _receivingSockets = <int, ChromeRawSocket>{};
  static bool isListeningChromeAPI = false;
  static void _startReceiving() {
    if (isListeningChromeAPI) {
      return;
    }
    isListeningChromeAPI = true;
    sockets.tcp.onReceiveError.listen((event) {});
    sockets.tcp.onReceive.listen((event) {
      final socketId = event.socketId;
      final socket = _receivingSockets[socketId];
      if (socket is ChromeRawSocket) {
        socket._incomingWriter.writeBytes(event.data.getBytes());
        socket._eventStreamController.add(RawSocketEvent.read);
      } else {
        print("Socket ID ${socketId} received data, but it's closed");
      }
    });
  }

  @override
  final int port;

  @override
  final InternetAddress address;

  @override
  StreamSubscription<RawSocketEvent> listen(void onData(RawSocketEvent event),
      {Function onError, void onDone(), bool cancelOnError}) {
    return _eventStreamController.stream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  @override
  bool readEventsEnabled;
  @override
  bool writeEventsEnabled;

  @override
  int available() {
    return _incomingWriter.length;
  }

  @override
  List<int> read([int len]) {
    final writer = this._incomingWriter;
    final bytes = writer.toUint8ListView();
    if (len == null || len == writer.length) {
      this._incomingWriter = RawWriter.withCapacity(64);
      return bytes;
    }
    final result = Uint8List.view(
      bytes.buffer,
      bytes.offsetInBytes,
      len,
    );
    final remaining = Uint8List.view(
      bytes.buffer,
      bytes.offsetInBytes + len,
    );
    this._incomingWriter = RawWriter.withUint8List(remaining);
    return result;
  }

  @override
  int write(List<int> buffer, [int offset, int count]) {
    if (offset != null || count != null) {
      offset ??= 0;
      count ??= buffer.length - offset;
      final newBuffer = Uint8List(count);
      for (var i = 0; i < count; i++) {
        newBuffer[i] = buffer[offset + i];
      }
      buffer = newBuffer;
    }
    sockets.tcp.send(socketId, ArrayBuffer.fromBytes(buffer));
    return count ?? buffer.length;
  }

  @override
  final int remotePort;

  @override
  final InternetAddress remoteAddress;

  @override
  Future<RawSocket> close() async {
    _receivingSockets.remove(socketId);
    sockets.tcp.close(socketId);
    return this;
  }

  @override
  void shutdown(SocketDirection direction) {
    if (direction == SocketDirection.both) {
      close();
    }
  }

  @override
  bool setOption(SocketOption option, bool enabled) {
    return false;
  }

  @override
  Uint8List getRawOption(RawSocketOption option) {
    return null;
  }

  @override
  void setRawOption(RawSocketOption option) {}
}

class ChromeRawDatagramSocket extends RawDatagramSocket {
  final int socketId;
  final StreamController<RawSocketEvent> eventStreamController =
      StreamController<RawSocketEvent>();
  Queue<Datagram> incomingQueue = Queue<Datagram>();

  ChromeRawDatagramSocket(this.socketId, {this.address, this.port});

  @override
  final int port;

  @override
  void leaveMulticast(InternetAddress group, [NetworkInterface interface]) {
    throw UnimplementedError();
  }

  @override
  Datagram receive() {
    return incomingQueue.removeFirst();
  }

  @override
  void close() {
    sockets.udp.close(socketId);
  }

  @override
  final InternetAddress address;

  @override
  StreamSubscription<RawSocketEvent> listen(void onData(RawSocketEvent event),
      {Function onError, void onDone(), bool cancelOnError}) {
    return eventStreamController.stream.listen(onData,
        onError: onError, onDone: onDone, cancelOnError: cancelOnError);
  }

  @override
  void joinMulticast(InternetAddress group, [NetworkInterface interface]) {
    throw UnimplementedError();
  }

  @override
  int send(List<int> buffer, InternetAddress address, int port) {
    sockets.udp.send(
        socketId, ArrayBuffer.fromBytes(buffer), address.toString(), port);
    return buffer.length;
  }

  @override
  void setRawOption(RawSocketOption option) {}

  @override
  Uint8List getRawOption(RawSocketOption option) {
    return null;
  }
}
